/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.sensor.greenseeker;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.EnumMap;
import java.util.List;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.SensorDataFile;
import nl.cloudfarming.client.model.SensorMetadata;
import nl.cloudfarming.client.sensor.SensorProcessor;
import nl.cloudfarming.client.sensor.MalformedFileFormatException;
import nl.cloudfarming.client.sensor.SensorEventKey;
import nl.cloudfarming.client.util.MD5Sum;
import nl.cloudfarming.eventbus.BaseGuiEventProducer;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.util.NbBundle;

/**
 * Greenseeker File processor
 * Processes normal greenseeker sensor files and v/d Borne Greenseeker sensorfiles.
 *
 * @author Gerben Feenstra
 */
public class GreenseekerFileProcessor extends SensorProcessor<GreenseekerSensorData> {

    // Messages
    private static final String LOGKEY_FAILED_TO_READ_FILE = "message.error.failed_to_read_file";
    private static final String LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG = "message.info.process_summary.processed";
    private static final String LOGKEY_PROCESSING_PARTIALLY_PROCESSEDMSG = "message.info.process_summary.partiallyprocessed";
    private static final String LOGKEY_PROCESSING_SUMMARY_ERRORMSG = "message.info.process_summary.errors";
    private static final String LOGKEY_PROCESSING_START = "message.info.process_start";
    // Constant int values
    private static final int HEADER_LENGTH_NORMAL = 481; // 672 = Header length Borne
    private static final int HEADER_LENGTH_BORNE = 672;
    private static final int RECORD_LENGTH_NORMAL = 211;
    private static final int RECORD_LENGTH_BORNE = 301;
    private static final int BULK_COMMIT_AMOUNT = 100; // How many record should be gathered before doing a bulk-commit?
    private static final int PROGRESS_REPORT_INTERVAL = 5000; // How many record should be processed before printing a progress message in the log

    // Date format variables
    private static final String PROPERTY_MONTH = "MM";
    private static final String PROPERTY_YEAR = "YYYY";
    private static final String PROPERTY_YEAR_INTERNATIONAL = "YYYY";
    private static final String PROPERTY_DAY = "DD";
    private static final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat("HHmmss.SS");

    // Other variable declarations
    private static final AppLogger LOG = AppLogFactory.getLogger(GreenseekerFileProcessor.class);
    private GreenseekerServiceImpl greenseekerService = new GreenseekerServiceImpl();
    private final ProgressHandle progressHandle;
    private final EventProducer eventProducer = new EventProducer();

    /**
     * Create a GreenseekerFileProcessor instance to process the specified greenseeker sensor file.
     *
     * @param toProcess The Greenseeker sensorfile to be processed
     */
    public GreenseekerFileProcessor(File toProcess) {
        super(toProcess);
        assert toProcess != null;
        this.progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(this.getClass(), "greenseeker.processor.processDescription", toProcess.getName()));
    }

    /**
     * Process the specified file
     */
    @Override
    public void process() {
        try {
            LOG.info(LOGKEY_PROCESSING_START, getProcessFile());

            // Create Datafile
            SensorMetadata metadata = buildMetadataFromPreferences();
            SensorDataFile dataFile = createSensorDataFile();
            greenseekerService.persist(dataFile);

            int success = 0;
            int fail = 0;
            int progress = 0;
            List<String> recordList = new ArrayList<String>();

            // Determine length + Create record byte[] container
            int recordLength = determineRecordLength();
            final byte[] record = new byte[recordLength];

            // Create new BufferedInputStream
            BufferedInputStream bis = new BufferedInputStream(new FileInputStream(getProcessFile()));

            // Skip header
            bis.skip(recordLength == RECORD_LENGTH_NORMAL ? HEADER_LENGTH_NORMAL : HEADER_LENGTH_BORNE);

            // Fill list
            while (bis.read(record) > 0) {
                recordList.add(new String(record));
            }

            List<GreenseekerSensorData> persistBuffer = new ArrayList<GreenseekerSensorData>();

            // Start progressbar
            this.progressHandle.start(recordList.size());

            // Start processing lines in list
            for (String line : recordList) {
                try {

                    GreenseekerSensorData toPersist;

                    if(recordLength == RECORD_LENGTH_BORNE){
                       toPersist = parseBorne(line, dataFile);
                    }
                    else {
                       toPersist = parseNormal(line, dataFile);
                    }

                    persistBuffer.add(toPersist);
                    success++;

                    // Persist
                    if (success % BULK_COMMIT_AMOUNT == 0) {
                        greenseekerService.persist(persistBuffer);
                        persistBuffer.clear();
                    }
                } catch (GreenseekerParseException ex) {
                    LOG.trace(ex);
                    fail++;
                } catch (Exception ex) {
                    LOG.trace(ex);
                    fail = fail + BULK_COMMIT_AMOUNT;
                    persistBuffer.clear();
                }

                progressHandle.progress(++progress);

                // Log partial progress at intervals
                if (progress % PROGRESS_REPORT_INTERVAL == 0) {
                    LOG.info(LOGKEY_PROCESSING_PARTIALLY_PROCESSEDMSG, progress);
                }
            }

            LOG.info(LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG, success);

            if(fail > 0){
                LOG.warning(LOGKEY_PROCESSING_SUMMARY_ERRORMSG, fail);
            }

            //Persist the remaining records
            greenseekerService.persist(persistBuffer);
            eventProducer.triggerEvent(SensorEventKey.SENSOR_DATA_IMPORTED);

            // Finish progressbar
            progressHandle.finish();
        } catch (Exception ex) {
            LOG.error(LOGKEY_FAILED_TO_READ_FILE, getProcessFile().getName());
            LOG.trace(ex);
        }

    }

    /**
     * Parses one record of a normal Greenseeker sensorfile.
     *
     * @param payload The record to be processed.
     * @param dataFile The SensorDataFile belonging to this sensor file.
     * @return A record represented as a Greenseeker sensordata object
     * @throws ParseException Exception occurs when date/time cannot be parsed.
     */
    private GreenseekerSensorData parseNormal(String payload, SensorDataFile dataFile) throws GreenseekerParseException {

        // Split payload into fields-array
        String[] fields = getFields(payload);

        // First create a entity representation of a file
        //converting the differrent data
        EnumMap map = new EnumMap<GreenseekerFileDescriptionNormal, Object>(GreenseekerFileDescriptionNormal.class);
        try{
            for (GreenseekerFileDescriptionNormal greenseekerNormalField : GreenseekerFileDescriptionNormal.values()) {
                if (greenseekerNormalField.isMandatory()) {
                    if (greenseekerNormalField.getTheClass().equals(java.lang.Double.class)) {
                        double theDouble = Double.parseDouble(fields[greenseekerNormalField.getPosition()]);
                        map.put(greenseekerNormalField, theDouble);
                    } else if (greenseekerNormalField.getTheClass().equals(java.lang.Integer.class)) {
                        int theInt = Integer.parseInt(fields[greenseekerNormalField.getPosition()]);
                        map.put(greenseekerNormalField, theInt);
                    } else if (greenseekerNormalField.getTheClass().equals(java.util.Date.class)) {
                        Date theDate = TIME_FORMAT.parse(fields[greenseekerNormalField.getPosition()]);
                        map.put(greenseekerNormalField, theDate);
                    } else {
                        map.put(greenseekerNormalField, fields[greenseekerNormalField.getPosition()]);
                    }
                }
            }
        }
        catch(ParseException e){
            throw new GreenseekerParseException("Error parsing date ", e);
        }
        catch(Exception e){
            throw new GreenseekerParseException(e);
        }

        GreenseekerSensorData sensorData = new GreenseekerSensorData(
                (SensorDataFile) dataFile,
                (Double) map.get(GreenseekerFileDescriptionNormal.OBJECTID),
                (Double) map.get(GreenseekerFileDescriptionNormal.LONGITUDE),
                (Double) map.get(GreenseekerFileDescriptionNormal.LATITUDE),
                (Double) map.get(GreenseekerFileDescriptionNormal.SPEED),
                (Double) map.get(GreenseekerFileDescriptionNormal.HEADING),
                (Double) map.get(GreenseekerFileDescriptionNormal.ELEVATION),
                (Double) map.get(GreenseekerFileDescriptionNormal.NDVI),
                (Double) map.get(GreenseekerFileDescriptionNormal.VI_2ND),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_10),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_19),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_23),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_24),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_26),
                (Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_73));
        return sensorData;
    }

    /**
     * Parses one record of a v/d Borne Greenseeker sensorfile.
     *
     * @param payload The record to be processed.
     * @param dataFile The SensorDataFile belonging to this sensor file.
     * @return A record represented as a Greenseeker sensordata object
     * @throws ParseException Exception occurs when date/time cannot be parsed.
     */
    private GreenseekerSensorData parseBorne(String payload, SensorDataFile dataFile) throws GreenseekerParseException {
        // Split payload into fields-array
        String[] fields = getFields(payload);

        EnumMap map = new EnumMap<GreenseekerFileDescriptionBorne, Object>(GreenseekerFileDescriptionBorne.class);
        try{
            for (GreenseekerFileDescriptionBorne greenseekerBorneField : GreenseekerFileDescriptionBorne.values()) {
                if (greenseekerBorneField.isMandatory()) {
                    if (greenseekerBorneField.getTheClass().equals(java.lang.Double.class)) {
                        double theDouble = Double.parseDouble(fields[greenseekerBorneField.getPosition()]);
                        map.put(greenseekerBorneField, theDouble);
                    } else if (greenseekerBorneField.getTheClass().equals(java.lang.Integer.class)) {
                        int theInt = Integer.parseInt(fields[greenseekerBorneField.getPosition()]);
                        map.put(greenseekerBorneField, theInt);
                    } else if (greenseekerBorneField.getTheClass().equals(java.util.Date.class)) {
                        Date theDate = TIME_FORMAT.parse(fields[greenseekerBorneField.getPosition()]);
                        map.put(greenseekerBorneField, theDate);
                    } else {
                        map.put(greenseekerBorneField, fields[greenseekerBorneField.getPosition()]);
                    }
                }
            }
        }
        catch(ParseException e){
            throw new GreenseekerParseException("Error parsing date ", e);
        }
        catch(Exception e){
            throw new GreenseekerParseException(e);
        }

        GreenseekerSensorData sensorData = new GreenseekerSensorData(
                (SensorDataFile) dataFile,
                (Double) map.get(GreenseekerFileDescriptionBorne.OBJECTID),
                (Double) map.get(GreenseekerFileDescriptionBorne.LONGITUDE),
                (Double) map.get(GreenseekerFileDescriptionBorne.LATITUDE),
                (Double) map.get(GreenseekerFileDescriptionBorne.SPEED),
                (Double) map.get(GreenseekerFileDescriptionBorne.HEADING),
                (Double) map.get(GreenseekerFileDescriptionBorne.ELEVATION),
                (Double) map.get(GreenseekerFileDescriptionBorne.NDVI),
                (Double) map.get(GreenseekerFileDescriptionBorne.VI2ND),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED54),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR54),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED55),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR55),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED57),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR57),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED53),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR53),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED56),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR56),
                (Double) map.get(GreenseekerFileDescriptionBorne.RED58),
                (Double) map.get(GreenseekerFileDescriptionBorne.NIR58));
        return sensorData;
    }

    /**
     * Extracts a date from the Greenseeker file name.
     *
     * @param fileName The filename to extract the date from
     * @return The extracted date
     * @throws MalformedFileFormatException Filename does not match the prescribed Greenseeker filenameformat.
     * @throws ParseException Exception occurs when the date cannot be parsed.
     */
    static Date getDateFromFileName(File sensorFile) throws MalformedFileFormatException, ParseException {
        String fileName = sensorFile.getName();
        String greenseekerDateFormat = GreenseekerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
        assert !greenseekerDateFormat.isEmpty();

        if (greenseekerDateFormat.contains(PROPERTY_DAY) && greenseekerDateFormat.contains(PROPERTY_MONTH) && greenseekerDateFormat.contains(PROPERTY_YEAR)) {
            int numberOfYearCharacters = PROPERTY_YEAR.length();
            if (greenseekerDateFormat.contains(PROPERTY_YEAR_INTERNATIONAL)) {
                numberOfYearCharacters = PROPERTY_YEAR_INTERNATIONAL.length();
            }
            String day = fileName.substring(greenseekerDateFormat.indexOf(PROPERTY_DAY), greenseekerDateFormat.indexOf(PROPERTY_DAY) + 2);
            String month = fileName.substring(greenseekerDateFormat.indexOf(PROPERTY_MONTH), greenseekerDateFormat.indexOf(PROPERTY_MONTH) + 2);
            String year = fileName.substring(greenseekerDateFormat.indexOf(PROPERTY_YEAR), greenseekerDateFormat.indexOf(PROPERTY_YEAR) + numberOfYearCharacters);
            StringBuilder dateBuilder = new StringBuilder(day);
            dateBuilder.append(month);
            dateBuilder.append(year);

            SimpleDateFormat date = new SimpleDateFormat("ddMMyy");
            return date.parse(dateBuilder.toString());
        } else {
            StringBuilder sb = new StringBuilder("The property which specifies file format is malformed while creating date from filename: ");
            sb.append(fileName);
            throw new MalformedFileFormatException(sb.toString());
        }
    }

    /**
     * Build Greenseeker metadata object
     *
     * @return Greenseeker metadata object
     */
    static SensorMetadata buildMetadataFromPreferences() {
        String fileFormat = GreenseekerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
        int sensorHeigth = Integer.parseInt(GreenseekerPreference.GREENSEEKER_SENSOR_HEIGTH.getValue());
        int numberOfSensors = Integer.parseInt(GreenseekerPreference.NUMBER_OF_GREENSEEKER_SENSORS.getValue());
        String greenseekerSensorSpread = GreenseekerPreference.GREENSEEKER_SENSOR_SPREAD.getValue();
        String sensorType = GreenseekerPreference.SENSOR_TYPE.getValue();
        String spectralChannels = GreenseekerPreference.SPECTRAL_CHANNELS.getValue();
        return new SensorMetadata(fileFormat, sensorHeigth, numberOfSensors, greenseekerSensorSpread, sensorType, spectralChannels);
    }

    /**
     * Creates the SensorDataFile object using the file that this fileprocessor should process
     * @return SensorDataFile object
     * @throws MalformedFileFormatException When filename (date-) format does not comply to greenseeker fileformat
     * @throws ParseException When filename (date-) format does not comply to greenseeker fileformat and cannot therefor not be parsed as a valid date
     * @throws NoSuchAlgorithmException When checksum fails
     * @throws IOException When the file this processor needs to process cannot be read.
     */
    private SensorDataFile createSensorDataFile() throws MalformedFileFormatException, ParseException, NoSuchAlgorithmException, IOException {
        SensorMetadata metadata = buildMetadataFromPreferences();
        Date runDate = getDateFromFileName(getProcessFile());
        String checkSum = MD5Sum.getHex(MD5Sum.createChecksum(new FileInputStream(getProcessFile())));
        String fileName = getProcessFile().getName();
        return new SensorDataFile(fileName, checkSum, runDate, new Date(), metadata);
    }

    /**
     * Determines the recordlength using the header of the file (to determine wether it is a normal Greenseeker sensorfile or a v/d Borne sensorfile
     *
     * @return recordlength
     * @throws FileNotFoundException When the file this processor needs to process cannot be found.
     * @throws IOException When the file this processor needs to process cannot be read.
     */
    private int determineRecordLength() throws IOException{
        // Start reading file
        BufferedInputStream bis = new BufferedInputStream(new FileInputStream(getProcessFile()));

        // Header normal= 481, borne = 672, use shortest to determine which filetype it is.
        final byte[] header = new byte[HEADER_LENGTH_NORMAL];
        bis.read(header);

        // Determine record length
        int recordLength = 0;

        // Recordlenght normal = 211, borne=301
        if (new String(header).contains("RED_54") && new String(header).contains("NIR_54")){
            recordLength = RECORD_LENGTH_BORNE;
        }
        else{
            recordLength = RECORD_LENGTH_NORMAL;
        }

        return recordLength;
    }

    /**
     * Removes multiple spaces from payload String and splits it into a String array
     *
     * @param payload The String to convert to an array
     * @return the splitted payload String
     */
    private String[] getFields(String payload){
        String payloadSpaced = payload.replaceAll("\\s{1,}", " ");
        String payloadTrimmed = payloadSpaced.trim();

        return payloadTrimmed.split(GreenseekerFileDescriptionBorne.getDelimiter());
    }

    @Override
    protected Date getRunDate(File sensorFile) throws MalformedFileFormatException, ParseException {
        throw new UnsupportedOperationException("Not supported yet.");
    }


    private class EventProducer extends BaseGuiEventProducer {

        public EventProducer() {
            super(SensorEventKey.class);
        }
    }
}
